"""
Based on unitests in https://github.com/wndhydrnt/python-oauth2/tree/master/oauth2/test
"""

import pytest
import unittest

from twitcher.datatype import AccessToken, Service
from twitcher import models
from twitcher.utils import expires_at


class AccessTokenTestCase(unittest.TestCase):

    def test_access_token(self):
        access_token = AccessToken(token='abcdef', expires_at=expires_at(hours=1))
        assert access_token.token == 'abcdef'
        assert access_token.expires_in > 0
        assert access_token.expires_in <= 3600
        assert access_token.is_expired() is False
        assert access_token.params['access_token'] == 'abcdef'
        assert 'expires_at' in access_token.params

    def test_missing_token(self):
        with pytest.raises(TypeError):
            AccessToken()

    def test_invalid_access_token(self):
        access_token = AccessToken(token='abcdef', expires_at=expires_at(hours=-1))
        assert access_token.expires_in == 0
        assert access_token.is_expired() is True

    def test_access_token_from_model(self):
        access_token = AccessToken.from_model(
            models.AccessToken(token='abc', expires_at=expires_at(hours=2)))
        assert access_token.token == 'abc'
        assert access_token.expires_at > 0


class ServiceTestCase(unittest.TestCase):
    def test_service_with_url_only(self):
        service = Service(name="test_wps", url='http://nowhere/wps')
        assert service.url == 'http://nowhere/wps'
        assert service.name == 'test_wps'
        assert service.has_purl() is False

    def test_missing_url_or_name(self):
        with pytest.raises(TypeError):
            Service(name="test")
        with pytest.raises(TypeError):
            Service(url='http://nowhere/wps')

    def test_service_with_name(self):
        service = Service(url='http://nowhere/wps', name="test_wps")
        assert service.url == 'http://nowhere/wps'
        assert service.name == 'test_wps'
        assert service.has_purl() is False

    def test_service_params(self):
        service = Service(url='http://nowhere/wps', name="test_wps", purl='http://myservice/wps')
        assert service.params == {'name': 'test_wps',
                                  'public': False,
                                  'auth': 'token',
                                  'type': 'WPS',
                                  'url': 'http://nowhere/wps',
                                  'verify': True,
                                  'purl': 'http://myservice/wps'}
        assert service.has_purl() is True

    def test_service_from_model(self):
        service = Service.from_model(
            models.Service(name='test_wps', url='http://nowhere/wps'))
        assert service.name == 'test_wps'
        assert service.url == 'http://nowhere/wps'
        assert service.type == 'WPS'
        assert service.auth == 'token'
        assert service.public is False
        assert service.verify is True
        assert service.purl == ''



from pyramid.config import Configurator


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    with Configurator(settings=settings) as config:
        # include twitcher components
        config.include('twitcher.models')
        config.include('twitcher.frontpage')
        config.include('twitcher.rpcinterface')
        config.include('twitcher.owsproxy')
        # tweens/middleware
        # TODO: maybe add tween for exception handling or use unknown_failure view
        config.include('twitcher.tweens')
        config.scan()
    return config.make_wsgi_app()



"""
Factories to create storage backends.
"""

from twitcher.adapter.base import AdapterInterface
from twitcher.store import AccessTokenStore, ServiceStore
from twitcher.owssecurity import OWSSecurity
from twitcher.utils import get_settings
from pyramid.config import Configurator

TWITCHER_ADAPTER_DEFAULT = 'default'


class DefaultAdapter(AdapterInterface):
    @property
    def name(self):
        return TWITCHER_ADAPTER_DEFAULT

    def describe_adapter(self):
        from twitcher.__version__ import __version__
        return {"name": self.name, "version": str(__version__)}

    def configurator_factory(self, container):
        settings = get_settings(container)
        return Configurator(settings=settings)

    def tokenstore_factory(self, request):
        return AccessTokenStore(request)

    def servicestore_factory(self, request):
        return ServiceStore(request)

    def owssecurity_factory(self, request):
        token_store = self.tokenstore_factory(request)
        service_store = self.servicestore_factory(request)
        return OWSSecurity(token_store, service_store)

    def owsproxy_config(self, container):
        from twitcher.owsproxy import owsproxy_defaultconfig
        # update provided config or generate it otherwise
        if not isinstance(container, Configurator):
            container = self.configurator_factory(container)
        owsproxy_defaultconfig(container)



from pyramid.settings import asbool
from pyramid.tweens import EXCVIEW

from twitcher.adapter import get_adapter_factory
from twitcher.owsexceptions import OWSException, OWSNoApplicableCode

import logging
LOGGER = logging.getLogger("TWITCHER")


def includeme(config):
    settings = config.registry.settings

    if asbool(settings.get('twitcher.ows_security', True)):
        LOGGER.info('Add OWS security tween')
        config.add_tween(OWS_SECURITY, under=EXCVIEW)


def ows_security_tween_factory(handler, registry):
    """A tween factory which produces a tween which raises an exception
    if access to OWS service is not allowed."""

    def ows_security_tween(request):
        try:
            adapter = get_adapter_factory(request)
            security = adapter.owssecurity_factory(request)
            security.check_request(request)
            return handler(request)
        except OWSException as err:
            LOGGER.warning("security check failed.")
            return err
        except Exception as err:
            LOGGER.exception("unknown error")
            return OWSNoApplicableCode("{}".format(err))

    return ows_security_tween


OWS_SECURITY = 'twitcher.tweens.ows_security_tween_factory'



"""
Read or write data from database.

Stores should not be accessed directly, but instead should use the adapter interface.

See also:
    - :class:`twitcher.adapter.base.AdapterInterface`
    - :func:`twitcher.adapter.get_adapter_factory`
"""

from twitcher.exceptions import (
    AccessTokenNotFound,
    ServiceNotFound,
    DatabaseError
)
from twitcher.utils import baseurl
from twitcher import datatype
from twitcher import models

from typing import TYPE_CHECKING
from sqlalchemy.exc import DBAPIError

if TYPE_CHECKING:
    from twitcher.models.token import AccessToken
    from twitcher.models.service import Service
    from pyramid.request import Request
    from typing import AnyStr, List


class AccessTokenStoreInterface(object):
    def __init__(self, request):            # type: (Request) -> None
        self.request = request

    def save_token(self, access_token):     # type: (AccessToken) -> None
        raise NotImplementedError

    def delete_token(self, token):          # type: (AccessToken) -> None
        raise NotImplementedError

    def fetch_by_token(self, token):        # type: (AnyStr) -> None
        raise NotImplementedError

    def clear_tokens(self):                 # type: () -> None
        raise NotImplementedError


class AccessTokenStore(AccessTokenStoreInterface):
    """
    Stores tokens in sql database.

    TODO: handle exceptions.
    """
    def __init__(self, request):
        super(AccessTokenStore, self).__init__(request)

    def save_token(self, access_token):
        """
        Stores an access token.
        """
        try:
            self.request.dbsession.add(models.AccessToken(
                token=access_token.token,
                expires_at=access_token.expires_at))
        except DBAPIError:
            raise DatabaseError

    def delete_token(self, token):
        """
        Deletes an access token from the store using its token string to identify it.

        :param token: A token string.
        """
        try:
            query = self.request.dbsession.query(models.AccessToken)
            one = query.filter(models.AccessToken.token == token).first()
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise AccessTokenNotFound
        self.request.dbsession.delete(one)

    def fetch_by_token(self, token):
        """
        Fetches an access token from the store using its token string to
        identify it.

        :param token: A token string.
        :return: An instance of :class:`twitcher.datatype.AccessToken`.
        """
        try:
            query = self.request.dbsession.query(models.AccessToken)
            one = query.filter(models.AccessToken.token == token).first()
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise AccessTokenNotFound
        return datatype.AccessToken.from_model(one)

    def clear_tokens(self):
        """
        Removes all tokens.
        """
        try:
            self.request.dbsession.query(models.AccessToken).delete()
        except DBAPIError:
            raise DatabaseError


class ServiceStoreInterface(object):
    def __init__(self, request):        # type: (Request) -> None
        self.request = request

    def save_service(self, service):    # type: (Service) -> None
        raise NotImplementedError

    def delete_service(self, name):     # type: (AnyStr) -> None
        raise NotImplementedError

    def list_services(self):            # type: () -> List[Service]
        raise NotImplementedError

    def fetch_by_name(self, name):      # type: (AnyStr) -> Service
        raise NotImplementedError

    def fetch_by_url(self, url):        # type: (AnyStr) -> Service
        raise NotImplementedError

    def clear_services(self):           # type: () -> None
        raise NotImplementedError


class ServiceStore(ServiceStoreInterface):
    """
    Stores a services. It inserts or updates the service with a given name.
    """
    def __init__(self, request):
        super(ServiceStore, self).__init__(request)

    def save_service(self, service):
        """
        Stores an OWS service in database (insert or update).

        :param service: An instance of :class:`twitcher.datatype.Service`.
        """
        try:
            query = self.request.dbsession.query(models.Service)
            one = query.filter(models.Service.name == service.name).first()
            if one:
                # update
                one.url = baseurl(service.url)
                one.type = service.type
                one.purl = service.purl
                one.public = int(service.public)
                one.verify = int(service.verify)
                one.auth = service.auth
                self.request.dbsession.merge(one)
            else:
                # insert
                self.request.dbsession.add(models.Service(
                    name=service.name,
                    url=baseurl(service.url),
                    type=service.type,
                    purl=service.purl,
                    public=int(service.public),
                    verify=int(service.verify),
                    auth=service.auth))
        except DBAPIError:
            raise DatabaseError

    def delete_service(self, name):
        """
        Removes service identified by name.
        """
        try:
            query = self.request.dbsession.query(models.Service)
            one = query.filter(models.Service.name == name).first()
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise ServiceNotFound
        self.request.dbsession.delete(one)

    def list_services(self):
        """
        Lists all services.

        :return: A list with instances of :class:`twitcher.datatype.Service`.
        """
        try:
            services = self.request.dbsession.query(models.Service).all()
        except DBAPIError:
            raise DatabaseError
        return [datatype.Service.from_model(service) for service in services]

    def fetch_by_name(self, name):
        """
        Get service for given service ``name``.

        :param name: A service name string.
        :return: An instance of :class:`twitcher.datatype.Service`.
        """
        try:
            query = self.request.dbsession.query(models.Service)
            one = query.filter(models.Service.name == name).first()
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise ServiceNotFound
        return datatype.Service.from_model(one)

    def fetch_by_url(self, url):
        """
        Get a service for given ``url``.

        :param url: A URL string.
        :return: An instance of :class:`twitcher.datatype.Service`.
        """
        try:
            query = self.request.dbsession.query(models.Service)
            one = query.filter(models.Service.url == url).first()
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise ServiceNotFound
        return datatype.Service.from_model(one)

    def clear_services(self):
        """
        Removes all services.
        """
        try:
            self.request.dbsession.query(models.Service).delete()
        except DBAPIError:
            raise DatabaseError



"""
Testing the Twitcher XML-RPC interface.
"""
import pytest

from .. common import WPS_TEST_SERVICE
from .base import FunctionalTest, call_FUT


class XMLRPCInterfaceAppTest(FunctionalTest):

    def setUp(self):
        super(XMLRPCInterfaceAppTest, self).setUp()
        self.init_database()
        self.init_store()

        self.config.include('twitcher.rpcinterface')
        self.app = self.test_app()

    def test_generate_token_and_revoke_it(self):
        # gentoken
        resp = call_FUT(self.app, 'generate_token', (1, ))
        assert 'access_token' in resp
        assert 'expires_at' in resp
        # revoke
        resp = call_FUT(self.app, 'revoke_token', (resp['access_token'],))
        assert resp is True
        # revoke all
        resp = call_FUT(self.app, 'revoke_all_tokens', ())
        assert resp is True

    def test_register_service_and_unregister_it(self):
        service = {'url': WPS_TEST_SERVICE, 'name': 'test_wps',
                   'type': 'wps', 'public': False, 'auth': 'token',
                   'verify': True, 'purl': 'http://purl/wps'}

        # clear
        resp = call_FUT(self.app, 'clear_services', ())
        assert resp is True

        # register
        resp = call_FUT(self.app, 'register_service', (
            service['name'],
            service['url'],
            service))
        assert resp == service

        # get by name
        resp = call_FUT(self.app, 'get_service_by_name', (service['name'],))
        assert resp == service

        # get by url
        resp = call_FUT(self.app, 'get_service_by_url', (service['url'],))
        assert resp == service

        # list
        resp = call_FUT(self.app, 'list_services', ())
        assert resp == [service]

        # unregister
        resp = call_FUT(self.app, 'unregister_service', (service['name'],))
        assert resp is True



import webtest

import xmlrpc.client as xmlrpclib

from .. common import BaseTest, WPS_TEST_SERVICE, dummy_request

from twitcher.utils import expires_at
from twitcher.datatype import AccessToken, Service
from twitcher.store import AccessTokenStore, ServiceStore

TEST_TOKEN = 'test_token_123'


def call_FUT(app, method, params):
    xml = xmlrpclib.dumps(params, methodname=method).encode('utf-8')
    print(xml)
    resp = app.post('/RPC2', content_type='text/xml', params=xml)
    assert resp.status_int == 200
    assert resp.content_type == 'text/xml'
    print(resp.body)
    return xmlrpclib.loads(resp.body)[0][0]


class FunctionalTest(BaseTest):
    def test_app(self):
        app = webtest.TestApp(
            self.config.make_wsgi_app(),
            extra_environ={'db.session': self.session, 'tm.active': True})
        return app

    def init_store(self):
        # add public wps service
        service_store = ServiceStore(
            dummy_request(dbsession=self.session))
        service_store.save_service(
            Service(
                name="wps",
                url=WPS_TEST_SERVICE,
                type="wps",
                public=True,
                auth='token',
                verify=False,
                purl="http://purl/wps"))
        # add secured wps service
        service_store.save_service(
            Service(
                name="wps_secured",
                url=WPS_TEST_SERVICE,
                type="wps",
                public=False,
                auth='token',
                verify=False,
                purl="http://purl/wps_secured"))
        # generate token
        token_store = AccessTokenStore(
            dummy_request(dbsession=self.session))
        token_store.save_token(
            AccessToken(token=TEST_TOKEN, expires_at=expires_at(hours=1)))



"""
Testing the Twitcher API.
"""

from .common import BaseTest, dummy_request

from twitcher.store import AccessTokenStore, ServiceStore
from twitcher.api import TokenManager, Registry
from twitcher.tokengenerator import UuidTokenGenerator


class TokenManagerTest(BaseTest):

    def setUp(self):
        super(TokenManagerTest, self).setUp()
        self.init_database()

        token_store = AccessTokenStore(
            dummy_request(dbsession=self.session))
        self.tokenmgr = TokenManager(
            tokengenerator=UuidTokenGenerator(),
            tokenstore=token_store
        )

    def test_generate_token_and_revoke_it(self):
        # gentoken
        resp = self.tokenmgr.generate_token()
        assert 'access_token' in resp
        assert 'expires_at' in resp
        # revoke
        resp = self.tokenmgr.revoke_token(resp['access_token'])
        assert resp is True
        # revoke all
        resp = self.tokenmgr.revoke_all_tokens()
        assert resp is True


class RegistryTest(BaseTest):

    def setUp(self):
        super(RegistryTest, self).setUp()
        self.init_database()

        service_store = ServiceStore(
            dummy_request(dbsession=self.session))

        self.reg = Registry(servicestore=service_store)

    def test_register_service_and_unregister_it(self):
        service = {'url': 'http://localhost/wps', 'name': 'test_emu',
                   'type': 'wps', 'public': False, 'auth': 'token', 'verify': True,
                   'purl': 'http://myservice/wps'}
        # register
        resp = self.reg.register_service(
            service['name'],
            service['url'],
            service)
        assert resp == service

        # get by name
        resp = self.reg.get_service_by_name(service['name'])
        assert resp == service

        # list
        resp = self.reg.list_services()
        assert resp == [service]

        # unregister
        resp = self.reg.unregister_service(service['name'])
        assert resp is True

        # clear
        resp = self.reg.clear_services()
        assert resp is True



import pytest

from pyramid.testing import DummyRequest, testConfig
from pyramid.testing import Registry

from .common import BaseTest, dummy_request

from twitcher.store import AccessTokenStore, ServiceStore
from twitcher.datatype import AccessToken, Service
from twitcher.utils import expires_at
from twitcher.owssecurity import OWSSecurity
from twitcher.owsexceptions import OWSAccessForbidden


class OWSSecurityTestCase(BaseTest):
    def setUp(self):
        super(OWSSecurityTestCase, self).setUp()
        self.init_database()

        request = dummy_request(dbsession=self.session)
        token_store = AccessTokenStore(request)
        token_store.save_token(AccessToken(token='test_token', expires_at=expires_at(hours=1)))
        service_store = ServiceStore(request)
        service_store.save_service(Service(name='test_wps', url='http://my/wps'))

        self.security = OWSSecurity(tokenstore=token_store, servicestore=service_store)

    def test_get_token_by_param(self):
        params = dict(request="Execute", service="WPS", access_token="abcdef")
        request = DummyRequest(params=params)
        token = self.security.get_token_param(request)
        assert token == "abcdef"

    def test_get_token_by_path(self):
        params = dict(request="Execute", service="WPS")
        request = DummyRequest(params=params, path="/ows/proxy/emu/12345")
        token = self.security.get_token_param(request)
        assert token == "12345"

    def test_get_token_by_header(self):
        params = dict(request="Execute", service="WPS")
        headers = {'Access-Token': '54321'}
        request = DummyRequest(params=params, headers=headers)
        token = self.security.get_token_param(request)
        assert token == "54321"

    def test_check_request(self):
        params = dict(request="Execute", service="WPS", version="1.0.0", token="test_token")
        request = DummyRequest(params=params, path='/ows/proxy/test_wps')
        request.registry = Registry()
        request.registry.settings = {'twitcher.ows_prox_protected_path': '/ows'}
        self.security.check_request(request)

    @pytest.mark.skip(reason="fix token check")
    def test_check_request_invalid(self):
        params = dict(request="Execute", service="WPS", version="1.0.0", token="invalid_token")
        request = DummyRequest(params=params, path='/ows/proxy/test_wps')
        request.registry = Registry()
        request.registry.settings = {'twitcher.ows_prox_protected_path': '/ows'}
        with pytest.raises(OWSAccessForbidden):
            self.security.check_request(request)

    def test_check_request_allowed_caps(self):
        params = dict(request="GetCapabilities", service="WPS", version="1.0.0")
        request = DummyRequest(params=params, path='/ows/proxy/test_wps')
        request.registry = Registry()
        request.registry.settings = {'twitcher.ows_prox_protected_path': '/ows'}
        self.security.check_request(request)

    def test_check_request_allowed_describeprocess(self):
        params = dict(request="DescribeProcess", service="WPS", version="1.0.0")
        request = DummyRequest(params=params, path='/ows/proxy/test_wps')
        request.registry = Registry()
        request.registry.settings = {'twitcher.ows_prox_protected_path': '/ows'}
        self.security.check_request(request)

    @pytest.mark.skip(reason="fix token check")
    def test_check_request_public_access(self):
        self.security.servicestore.save_service(Service(
            url='http://nowhere/wps', name='new_wps', public=True))
        params = dict(request="Execute", service="WPS", version="1.0.0", token="invalid_token")
        request = DummyRequest(params=params, path='/ows/proxy/new_wps')
        request.registry = Registry()
        request.registry.settings = {'twitcher.ows_prox_protected_path': '/ows'}
        self.security.check_request(request)



from sqlalchemy import (
    Column,
    Index,
    Integer,
    Text,
)

from .meta import Base


class AccessToken(Base):
    __tablename__ = 'access_tokens'
    id = Column(Integer, primary_key=True)
    token = Column(Text)
    expires_at = Column(Integer)


Index('token_index', AccessToken.token, unique=True, mysql_length=255)



from sqlalchemy import (
    Column,
    Index,
    Integer,
    Text,
)

from .meta import Base


class Service(Base):
    __tablename__ = 'services'
    id = Column(Integer, primary_key=True)
    url = Column(Text)
    name = Column(Text)
    type = Column(Text)
    purl = Column(Text)
    public = Column(Integer)  # sqlite does not support Boolean
    verify = Column(Integer)  # sqlite does not support Boolean
    auth = Column(Text)


Index('name_index', Service.name, unique=True, mysql_length=255)



"""
Definition of service and token types.
"""

import time

from pyramid.settings import asbool

from twitcher.utils import now_secs, is_valid_url
from twitcher.exceptions import AccessTokenNotFound


class Service(dict):
    """
    Dictionary that contains OWS services.
    It always has the ``'url'`` and ``'name'`` key.

    TODO: Should this be part of the model?
    TODO: If the field is accessed with service['public'] instead of service.public,
    we get different values. We could override `__getattr__`.
    """
    def __init__(self, *args, **kwargs):
        super(Service, self).__init__(*args, **kwargs)
        if 'name' not in self:
            raise TypeError("'name' is required")
        if 'url' not in self:
            raise TypeError("'url' is required")
        if self.get('public') is None:
            self['public'] = False
        if self.get('verify') is None:
            self['verify'] = True

    @property
    def url(self):
        """Service URL."""
        return self['url']

    @property
    def name(self):
        """Service name."""
        return self['name']

    @property
    def type(self):
        """Service type."""
        return self.get('type') or 'WPS'

    @property
    def purl(self):
        """Service optional public URL (purl)."""
        return self.get('purl') or ''

    def has_purl(self):
        """Return true if we have a valid public URL (purl)."""
        return is_valid_url(self.purl)

    @property
    def public(self):
        """Flag if service has public access."""
        # TODO: public access can be set via auth parameter.
        return asbool(self.get('public', False))

    @property
    def auth(self):
        """Authentication method: public, token, cert."""
        return self.get('auth') or 'token'

    @property
    def verify(self):
        """Verify ssl service certificate."""
        return asbool(self.get('verify', True))

    @property
    def params(self):
        return {
            'url': self.url,
            'name': self.name,
            'type': self.type,
            'purl': self.purl,
            'public': self.public,
            'auth': self.auth,
            'verify': self.verify}

    @classmethod
    def from_model(cls, model):
        instance = cls(
            url=model.url,
            name=model.name,
            type=model.type,
            purl=model.purl,
            public=model.public,
            auth=model.auth,
            verify=model.verify)
        return instance

    def __str__(self):
        return self.name

    def __repr__(self):
        cls = type(self)
        repr_ = dict.__repr__(self)
        return '{0}.{1}({2})'.format(cls.__module__, cls.__name__, repr_)


class AccessToken(dict):
    """
    Dictionary that contains access token. It always has ``'token'`` key.

    TODO: Should this be part of the model?
    """

    def __init__(self, *args, **kwargs):
        super(AccessToken, self).__init__(*args, **kwargs)
        if 'token' not in self:
            raise TypeError("'token' is required")

    @property
    def token(self):
        """Access token string."""
        return self['token']

    @property
    def expires_at(self):
        return int(self.get("expires_at") or 0)

    @property
    def expires_in(self):
        """
        Returns the time until the token expires.
        :return: The remaining time until expiration in seconds or 0 if the
                 token has expired.
        """
        time_left = self.expires_at - now_secs()

        if time_left > 0:
            return time_left
        return 0

    def is_expired(self):
        """
        Determines if the token has expired.
        :return: `True` if the token has expired. Otherwise `False`.
        """
        if self.expires_at is None:
            return True

        if self.expires_in > 0:
            return False

        return True

    @property
    def data(self):
        return self.get('data') or {}

    @property
    def params(self):
        return {'access_token': self.token, 'expires_at': self.expires_at}

    @classmethod
    def from_model(cls, model):
        instance = cls(
            token=model.token,
            expires_at=model.expires_at)
        return instance

    def __str__(self):
        return self.token

    def __repr__(self):
        cls = type(self)
        repr_ = dict.__repr__(self)
        return '{0}.{1}({2})'.format(cls.__module__, cls.__name__, repr_)



from twitcher.exceptions import AccessTokenNotFound, ServiceNotFound
from twitcher.owsexceptions import OWSAccessForbidden, OWSInvalidParameterValue
from twitcher.utils import path_elements, parse_service_name
from twitcher.owsrequest import OWSRequest
from pyramid.httpexceptions import HTTPNotFound

import logging
LOGGER = logging.getLogger("TWITCHER")


def verify_cert(request):
    if not request.headers.get('X-Ssl-Client-Verify', '') == 'SUCCESS':
        raise OWSAccessForbidden("A valid X.509 client certificate is needed.")


class OWSSecurityInterface(object):

    def check_request(self, request):
        raise NotImplementedError


class OWSSecurity(OWSSecurityInterface):

    def __init__(self, tokenstore, servicestore):
        self.tokenstore = tokenstore
        self.servicestore = servicestore

    @staticmethod
    def get_token_param(request):
        token = None
        if 'token' in request.params:
            token = request.params['token']   # in params
        elif 'access_token' in request.params:
            token = request.params['access_token']   # in params
        elif 'Access-Token' in request.headers:
            token = request.headers['Access-Token']  # in header
        else:  # in path
            elements = path_elements(request.path)
            if len(elements) > 1:  # there is always /ows/
                token = elements[-1]   # last path element
        return token

    def verify_access(self, request, service):
        # TODO: public service access handling is confusing.
        try:
            if service.auth == 'cert':
                verify_cert(request)
            else:  # token
                self._verify_access_token(request)
        except OWSAccessForbidden:
            if not service.public:
                raise

    def _verify_access_token(self, request):
        try:
            # try to get access_token ... if no access restrictions then don't complain.
            token = self.get_token_param(request)
            access_token = self.tokenstore.fetch_by_token(token)
            if access_token.is_expired():
                raise OWSAccessForbidden("Access token is expired.")
        except AccessTokenNotFound:
            raise OWSAccessForbidden("Access token is required to access this service.")

    def check_request(self, request):
        protected_path = request.registry.settings.get('twitcher.ows_proxy_protected_path ', '/ows')
        if request.path.startswith(protected_path):
            # TODO: refactor this code
            try:
                service_name = parse_service_name(request.path, protected_path)
                service = self.servicestore.fetch_by_name(service_name)
                if service.public is True:
                    LOGGER.warning('public access for service %s', service_name)
            except ServiceNotFound:
                raise OWSInvalidParameterValue(
                    "Service not found", value="service", status_base=HTTPNotFound)
            ows_request = OWSRequest(request)
            if not ows_request.service_allowed():
                raise OWSInvalidParameterValue(
                    "Service {} not supported".format(ows_request.service), value="service")
            if not ows_request.public_access():
                self.verify_access(request, service)
